import os
from pathlib import Path

import cv2
import numpy as np

from neurovc.util import normalize_color

try:
    import torch

    HAS_TORCH = True
except ModuleNotFoundError:
    torch = None
    HAS_TORCH = False

try:
    import yolov5_face.detect_face as yf

    HAS_YOLOV5_FACE = True
except ModuleNotFoundError:
    yf = None
    HAS_YOLOV5_FACE = False

try:
    import requests

    HAS_REQUESTS = True
except ModuleNotFoundError:
    requests = None
    HAS_REQUESTS = False

try:
    from tqdm import tqdm

    HAS_TQDM = True
except ModuleNotFoundError:
    tqdm = None
    HAS_TQDM = False


class _ModelDownloader:
    def __init__(self, model_name, save_dir="~/.neurovc/models"):
        if not HAS_REQUESTS:
            raise ImportError(
                "requests is required to download models. Install the landmark extra: 'pip install neurovc[landmark]'"
            )
        self.model_name = model_name
        self.save_dir = Path(os.path.expanduser(save_dir))
        self.file_id = _file_id_map.get(model_name)
        if not self.file_id:
            raise ValueError(
                f"Model name '{model_name}' is not valid. Check the file_id_map."
            )
        self.model_url = (
            f"https://drive.google.com/uc?export=download&id={self.file_id}"
        )
        self.model_path = self.save_dir / f"{model_name}.pt"

    def download_model(self):
        self.save_dir.mkdir(parents=True, exist_ok=True)

        if self.model_path.exists():
            return self.model_path

        print(f"Downloading {self.model_name}...")
        response = requests.get(self.model_url, stream=True)
        if response.status_code == 200:
            total_size = int(response.headers.get("content-length", 0))
            progress_iter = None
            if HAS_TQDM and total_size:
                progress_iter = tqdm(
                    desc=f"Downloading {self.model_name}",
                    total=total_size,
                    unit="B",
                    unit_scale=True,
                    unit_divisor=1024,
                )
            with open(self.model_path, "wb") as f:
                for chunk in response.iter_content(chunk_size=8192):
                    f.write(chunk)
                    if progress_iter is not None:
                        progress_iter.update(len(chunk))
            if progress_iter is not None:
                progress_iter.close()
            print(f"Model downloaded to {self.model_path}")
        else:
            raise Exception(f"Failed to download model: {response.status_code}")
        return self.model_path


def _prepare_model(model_name="YOLOv5n-Face"):
    downloader = _ModelDownloader(model_name)
    model_path = downloader.download_model()
    return model_path


class LandmarkWrapper:
    def process(self, img):
        return self.get_landmarks(img), None


class TFWLandmarker(LandmarkWrapper):
    def __init__(self, model_name="YOLOv5n-Face"):
        if not HAS_TORCH:
            raise ImportError(
                "torch is required for TFWLandmarker. Install the torch extra: 'pip install neurovc[torch]'"
            )
        if not HAS_YOLOV5_FACE:
            raise ImportError(
                "yolov5-face is required for TFWLandmarker. Install the landmark extra: 'pip install neurovc[landmark]'"
            )
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        model_path = _prepare_model(model_name)
        self.model = yf.load_model(model_path, self.device)

    def detect(self, img):
        img = normalize_color(img, color_map=cv2.COLORMAP_BONE)
        results = yf.detect_landmarks(self.model, img, self.device)
        return results

    def get_landmarks(self, img):
        results = self.detect(img)
        if len(results) == 0:
            return np.full((5, 2), -1)
        lm = results[0]["landmarks"]
        lm = np.array(lm).reshape((-1, 2))
        return lm


_file_id_map = {
    "YOLOv5n": "1PLUq7WbOWS7Ve2VKW7_WBkC3Uksje8Fx",
    "YOLOv5n6": "1wV9t5uH_eiy7WaHdQdWnbeEIijuDAdKI",
    "YOLOv5s": "1IdsdR1-qUeRo5EKQJzGQmRDi2SrMXJG5",
    "YOLOv5s6": "1YZX3t7cSPnWWoic7oJo86ljBQgE5PPb2",
    "YOLOv5n-Face": "1vXk9P3CfhUtRBGI44SqWbuiTJ7rAI4hP",
}


__all__ = [
    "HAS_TORCH",
    "HAS_YOLOV5_FACE",
    "HAS_REQUESTS",
    "HAS_TQDM",
    "TFWLandmarker",
    "LandmarkWrapper",
]



from . import tfw_util as _tfw_util

HAS_TORCH = _tfw_util.HAS_TORCH
HAS_YOLOV5_FACE = _tfw_util.HAS_YOLOV5_FACE
HAS_REQUESTS = _tfw_util.HAS_REQUESTS
HAS_TQDM = _tfw_util.HAS_TQDM

__all__ = [
    "HAS_TORCH",
    "HAS_YOLOV5_FACE",
    "HAS_REQUESTS",
    "HAS_TQDM",
    *_tfw_util.__all__,
]

for _name in _tfw_util.__all__:
    globals()[_name] = getattr(_tfw_util, _name)

del _tfw_util, _name



from neurovc.util.IO_util import *


def __preproc1(frame_cel):
    frame_cel[frame_cel < 20] = 20
    frame_cel[frame_cel > 45] = 45
    frame_cel = frame_cel + 2 * (frame_cel - imgaussfilt(frame_cel, 5))
    return cv2.normalize(frame_cel, None, alpha=0, beta=255, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_8U)


def __preproc2(frame_cel):
    frame_cel[frame_cel < 20] = 20
    frame_cel[frame_cel > 45] = 45
    frame_cel = frame_cel + 4 * (frame_cel - cv2.GaussianBlur(frame_cel, (15, 15), 5))
    return cv2.normalize(frame_cel, None, alpha=0, beta=255, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_8U)


def __preproc3(frame_cel):
    return 1 - __preproc1(frame_cel)


def __preproc4(frame_cel):
    return 1 - __preproc2(frame_cel)


def get_preprocessing_stack():
    return [__preproc1, __preproc2, __preproc3, __preproc4]


"""Utilities for spatial warping and splatting in the motion magnifier."""

__author__ = "Cosmas Heiss, Philipp Flotho"

try:
    import moderngl

    HAS_MODERNGL = True
except ModuleNotFoundError:
    moderngl = None
    HAS_MODERNGL = False

try:
    import torch

    HAS_TORCH = True
except ModuleNotFoundError:
    torch = None
    HAS_TORCH = False

import numpy as np
from scipy.interpolate import griddata
import cv2
from typing import Tuple


def warp_image_pc(img, flow):
    """Warp an RGB image using SciPy point-cloud interpolation.

    Parameters
    ----------
    img : ndarray
        Source image with shape ``(H, W, 3)``.
    flow : ndarray
        Forward flow with shape ``(H, W, 2)`` in pixel coordinates.

    Returns
    -------
    ndarray
        Warped RGB image with the same shape and dtype as ``img``.
    """
    m, n = img.shape[:2]
    xi, yi = np.meshgrid(np.arange(n).astype(float), np.arange(m).astype(float))

    tmp = np.empty(img.shape, img.dtype)
    for i in range(3):
        griddata_result = griddata(
            (
                xi.flatten() + flow[:, :, 0].flatten(),
                yi.flatten() + flow[:, :, 1].flatten(),
            ),
            img[:, :, i].flatten(),
            (xi.flatten(), yi.flatten()),
            method="linear",
            fill_value=0,
        )
        tmp[:, :, i] = np.reshape(griddata_result, (m, n))

    return tmp


def warp_image_pc_single(img, flow):
    """Warp a single-channel image using point-cloud interpolation.

    Parameters
    ----------
    img : ndarray
        Source image with shape ``(H, W)``.
    flow : ndarray
        Forward flow with shape ``(H, W, 2)`` in pixel coordinates.

    Returns
    -------
    ndarray
        Warped image with the same shape and dtype as ``img``.
    """
    m, n = img.shape[:2]
    xi, yi = np.meshgrid(np.arange(n).astype(float), np.arange(m).astype(float))

    tmp = np.empty(img.shape, img.dtype)

    griddata_result = griddata(
        (
            xi.flatten() + flow[:, :, 0].flatten(),
            yi.flatten() + flow[:, :, 1].flatten(),
        ),
        img[:, :].flatten(),
        (xi.flatten(), yi.flatten()),
        method="nearest",
        fill_value=0,
    )
    tmp[:, :] = np.reshape(griddata_result, (m, n))

    return tmp


def warp_image_backwards(img, flow):
    """Backward warp an image using OpenCV's remap interface.

    Parameters
    ----------
    img : ndarray
        Source image with shape ``(H, W, C)``.
    flow : ndarray
        Backward flow with shape ``(H, W, 2)`` mapping output pixels to source
        locations.

    Returns
    -------
    ndarray
        Backward-warped image with the same shape and dtype as ``img``.
    """
    h, w = flow.shape[:2]
    # flow = -flow
    tmp_flow = np.empty(flow.shape, np.array(flow).dtype)
    np.copyto(tmp_flow, flow)
    tmp_flow[:, :, 0] += np.arange(w)
    tmp_flow[:, :, 1] += np.arange(h)[:, np.newaxis]
    res = cv2.remap(img, tmp_flow, None, cv2.INTER_LINEAR)
    return res


# moderngl-based fast forward warping, author: Cosmas Heiß
class OnlineFrameWarper:
    """Moderngl-backed forward warper for dense meshes."""

    def __init__(self, image_size):
        """Initialise the warper and allocate moderngl buffers.

        Parameters
        ----------
        image_size : tuple of int
            Output image size as ``(height, width)``.

        Raises
        ------
        ImportError
            If ``moderngl`` is not installed.
        """
        if not HAS_MODERNGL:
            raise ImportError(
                "moderngl is required for OnlineFrameWarper. Install the gui extra: 'pip install neurovc[gui]'"
            )
        self.image_size = image_size
        self.strip_indices = self.generate_triangle_strip_index_array()
        self.ctx = moderngl.create_context(standalone=True, require=330)
        self.ctx.enable(moderngl.DEPTH_TEST)

        self.prog = self.ctx.program(
            vertex_shader="""
                #version 330

                in vec2 in_vert;
                in vec3 in_color;
                in float in_depth;
                out vec3 v_color;


                void main() {
                    v_color = in_color;
                    gl_Position = vec4(in_vert, in_depth, 1.0);
                }
            """,
            fragment_shader="""
                #version 330

                in vec3 v_color;

                out vec3 f_color;

                void main() {
                    f_color = v_color;
                }
            """,
        )
        dummy_vertices = self.get_dummy_vertices()
        self.vertex_buffer = self.ctx.buffer(dummy_vertices)
        self.vertex_array = self.ctx.vertex_array(
            self.prog, self.vertex_buffer, "in_vert", "in_color", "in_depth"
        )
        self.frame_buffer = self.ctx.simple_framebuffer(image_size[::-1])

    def get_dummy_vertices(self):
        """Return a dummy vertex buffer covering the full image plane.

        Returns
        -------
        bytes
            Vertex buffer encoded as 32-bit floats arranged in
            ``(x, y, r, g, b, depth)`` order.
        """
        image = np.zeros((*self.image_size, 3))
        xx, yy = np.meshgrid(
            np.linspace(-1, 1, self.image_size[1], endpoint=True),
            np.linspace(-1, 1, self.image_size[0], endpoint=True),
        )
        displacements = np.stack((xx, yy), axis=2)
        depth = np.zeros(self.image_size)
        return self.get_into_vertex_buffer_shape(image, displacements, depth)

    def vertices_astype(self, vertices):
        """Convert vertex arrays to 32-bit floating point byte buffers.

        Parameters
        ----------
        vertices : ndarray
            Vertex array to serialise.

        Returns
        -------
        bytes
            Serialised vertex buffer.
        """
        return vertices.astype("f4").tobytes()

    def get_into_vertex_buffer_shape(self, image, displacements, depth):
        """Pack image, displacement, and depth channels into vertex layout.

        Parameters
        ----------
        image : ndarray
            RGB image normalised to ``[0, 1]`` with shape ``(H, W, 3)``.
        displacements : ndarray
            Flow field with shape ``(H, W, 2)``.
        depth : ndarray
            Depth buffer with shape ``(H, W)``.

        Returns
        -------
        bytes
            Serialised vertex buffer for the triangle strip.
        """
        vertices = np.concatenate((displacements, image, depth[:, :, None]), axis=2)[
            self.strip_indices[:, 0], self.strip_indices[:, 1]
        ]
        return self.vertices_astype(vertices)
        # return self.get_into_vb_fast(image, displacements, depth)

    # def get_into_vb_fast(self, image, displacements, depth):
    #    image = cp.asarray(image)
    #    displacements = cp.asarray(displacements)
    #    depth = cp.asarray(depth)
    #    strip_indices = cp.asarray(self.strip_indices)
    #    vertices = cp.concatenate((displacements, image, depth[:, :, None]), axis=2)[strip_indices[:, 0], strip_indices[:, 1]]
    #    return self.vertices_astype(cp.asnumpy(vertices))

    def generate_triangle_strip_index_array(self):
        """Generate a scan-line triangle strip index buffer.

        Returns
        -------
        ndarray
            Array of indices with shape ``(N, 2)`` describing the strip order.
        """
        out_indices_x = []
        out_indices_y = []
        for i in range(self.image_size[1] - 1):
            is_reversed = int((i % 2) * (-2) + 1)
            out_indices_x.append(np.arange(self.image_size[0]).repeat(2)[::is_reversed])
            out_indices_y.append(np.tile(np.array([i, i + 1]), self.image_size[0]))

        out_indices_x = np.concatenate(out_indices_x).astype(int)
        out_indices_y = np.concatenate(out_indices_y).astype(int)
        return np.stack((out_indices_x, out_indices_y), axis=1)

    def read_frame_buffer(self):
        """Return the framebuffer contents as ``uint8`` pixels.

        Returns
        -------
        ndarray
            Array containing the framebuffer data with shape ``(H, W, 4)``.
        """
        return np.frombuffer(self.frame_buffer.read(), "uint8").reshape(
            self.image_size[0], self.image_size[1], -1
        )

    def pixel_to_screenspace_coords(self, displacements):
        """Map pixel displacements to OpenGL clip-space coordinates.

        Parameters
        ----------
        displacements : ndarray
            Flow field with shape ``(H, W, 2)`` in pixel units.

        Returns
        -------
        ndarray
            Displacements mapped to clip space with the same shape as input.
        """
        out = np.zeros_like(displacements)
        out[:, :, 1] = displacements[:, :, 0] * (2.0 / (self.image_size[0] - 1)) - 1.0
        out[:, :, 0] = displacements[:, :, 1] * (2.0 / (self.image_size[1] - 1)) - 1.0
        return out

    def warp_image(self, image, displacements, depth):
        """Render an image under forward displacements and depth ordering.

        Parameters
        ----------
        image : ndarray
            Normalised RGB image in ``[0, 1]`` with shape ``(H, W, 3)``.
        displacements : ndarray
            Flow field with shape ``(H, W, 2)`` in pixel units.
        depth : ndarray
            Depth buffer with shape ``(H, W)``.

        Returns
        -------
        ndarray
            Rendered RGB image with shape ``(H, W, 3)`` and dtype ``float64``.
        """
        assert image.dtype == displacements.dtype == depth.dtype == float
        assert np.all(np.logical_and(image <= 1.0, image >= 0.0))
        assert (
            image.shape[:2] == displacements.shape[:2] == depth.shape == self.image_size
        )

        displacements = self.pixel_to_screenspace_coords(displacements)
        if depth.max() != depth.min():
            depth = -0.99 * (depth - depth.min()) / (depth.max() - depth.min())

        self.frame_buffer.use()
        self.frame_buffer.clear(0.0, 0.0, 0.0, 0.0)

        self.vertex_buffer.write(
            self.get_into_vertex_buffer_shape(image, displacements, depth)
        )

        self.vertex_array.render(moderngl.TRIANGLE_STRIP)

        return self.read_frame_buffer()[:, :, :3]

    def warp_image_uv(self, image, uv, depth=None):
        """Warp an image using UV flow (``u`` = x, ``v`` = y) and optional depth.

        Parameters
        ----------
        image : ndarray
            Source RGB image with shape ``(H, W, 3)``.
        uv : ndarray
            Flow field with shape ``(H, W, 2)`` where ``uv[..., 0]`` and
            ``uv[..., 1]`` represent ``u`` and ``v`` displacements.
        depth : ndarray, optional
            Depth buffer with shape ``(H, W)``. If omitted, zeros are used.

        Returns
        -------
        ndarray
            Rendered RGB image with shape ``(H, W, 3)``.
        """
        m, n = uv.shape[0:2]

        x, y = np.meshgrid(np.arange(n).astype(float), np.arange(m).astype(float))

        tmp_flow = np.empty(uv.shape, float)
        tmp_flow[:, :, 0] = y + uv[:, :, 1]
        tmp_flow[:, :, 1] = x + uv[:, :, 0]

        if depth is None:
            depth = np.zeros((m, n), float)

        return self.warp_image(np.array(image).astype(float) / 255.0, tmp_flow, depth)


def _to_bchw(image: "torch.Tensor") -> Tuple["torch.Tensor", str]:
    """Convert an image tensor to ``(B, C, H, W)`` layout.

    Parameters
    ----------
    image : torch.Tensor
        Image tensor with 2–4 dimensions arranged as either
        ``(H, W)``, ``(C, H, W)``, ``(H, W, C)``, or their batched variants.

    Returns
    -------
    torch.Tensor
        Tensor reshaped to ``(B, C, H, W)``.
    str
        Identifier describing the original layout so it can be restored later.

    Raises
    ------
    ValueError
        If the layout cannot be inferred unambiguously.
    """
    tensor = image
    if tensor.ndim == 4:
        if tensor.shape[1] <= 4:
            layout = "bchw"
        elif tensor.shape[-1] <= 4:
            tensor = tensor.permute(0, 3, 1, 2)
            layout = "bhwc"
        else:
            raise ValueError("Expected channels-first or channels-last 4D tensor.")
    elif tensor.ndim == 3:
        if tensor.shape[0] <= 4 and tensor.shape[0] != tensor.shape[2]:
            tensor = tensor.unsqueeze(0)
            layout = "chw"
        elif tensor.shape[-1] <= 4:
            tensor = tensor.permute(2, 0, 1).unsqueeze(0)
            layout = "hwc"
        else:
            raise ValueError(
                "Ambiguous 3D tensor layout; cannot infer channel dimension."
            )
    elif tensor.ndim == 2:
        tensor = tensor.unsqueeze(0).unsqueeze(0)
        layout = "hw"
    else:
        raise ValueError("Expected tensor with 2 to 4 dimensions.")
    return tensor, layout


def _from_bchw(tensor: "torch.Tensor", layout: str):
    """Restore a tensor to its original layout after :func:`_to_bchw`.

    Parameters
    ----------
    tensor : torch.Tensor
        Tensor in ``(B, C, H, W)`` layout.
    layout : str
        Layout identifier returned by :func:`_to_bchw`.

    Returns
    -------
    torch.Tensor
        Tensor reshaped back to its original layout.

    Raises
    ------
    ValueError
        If ``layout`` is unknown.
    """
    if layout == "bchw":
        return tensor
    if layout == "bhwc":
        return tensor.permute(0, 2, 3, 1)
    if layout == "chw":
        return tensor.squeeze(0)
    if layout == "hwc":
        return tensor.squeeze(0).permute(1, 2, 0)
    if layout == "hw":
        return tensor.squeeze(0).squeeze(0)
    raise ValueError(f"Unknown layout identifier '{layout}'.")


def _to_b2hw(flow: "torch.Tensor") -> Tuple["torch.Tensor", str]:
    """Convert a flow tensor to ``(B, 2, H, W)`` layout.

    Parameters
    ----------
    flow : torch.Tensor
        Flow tensor with 3–4 dimensions arranged as either
        ``(2, H, W)``, ``(H, W, 2)``, or their batched variants.

    Returns
    -------
    torch.Tensor
        Tensor reshaped to ``(B, 2, H, W)``.
    str
        Identifier describing the original layout.

    Raises
    ------
    ValueError
        If the layout cannot be inferred.
    """
    tensor = flow
    if tensor.ndim == 4:
        if tensor.shape[1] == 2:
            layout = "b2hw"
        elif tensor.shape[-1] == 2:
            tensor = tensor.permute(0, 3, 1, 2)
            layout = "bhw2"
        else:
            raise ValueError("Expected flow tensor with 2 channels.")
    elif tensor.ndim == 3:
        if tensor.shape[0] == 2:
            tensor = tensor.unsqueeze(0)
            layout = "2hw"
        elif tensor.shape[-1] == 2:
            tensor = tensor.permute(2, 0, 1).unsqueeze(0)
            layout = "hw2"
        else:
            raise ValueError("Expected flow tensor with 2 components.")
    else:
        raise ValueError("Flow tensor must have 3 or 4 dimensions.")
    return tensor, layout


def _prepare_spatial_map(
    value,
    batch: int,
    height: int,
    width: int,
    device: "torch.device",
    dtype: "torch.dtype",
):
    """Broadcast a spatial map to ``(B, 1, H, W)``.

    Parameters
    ----------
    value : array-like or torch.Tensor or None
        Spatial weighting map. ``None`` results in a tensor of ones.
    batch : int
        Target batch size.
    height : int
        Target image height.
    width : int
        Target image width.
    device : torch.device
        Target device.
    dtype : torch.dtype
        Target dtype.

    Returns
    -------
    torch.Tensor
        Broadcasted tensor with shape ``(B, 1, H, W)``.

    Raises
    ------
    ValueError
        If ``value`` cannot be broadcast to the requested shape.
    """
    if value is None:
        return torch.ones((batch, 1, height, width), device=device, dtype=dtype)

    tensor = value if torch.is_tensor(value) else torch.as_tensor(value, device=device)
    tensor = tensor.to(device=device, dtype=dtype)

    if tensor.ndim == 4:
        if tensor.shape[0] not in (1, batch):
            raise ValueError("Spatial map batch dimension does not match image batch.")
        if tensor.shape[0] == 1 and batch != 1:
            tensor = tensor.expand(batch, *tensor.shape[1:])
        if tensor.shape[1] == 1:
            return tensor.contiguous()
        if tensor.shape[-1] == 1:
            return tensor.permute(0, 3, 1, 2).contiguous()
    elif tensor.ndim == 3:
        if tensor.shape[0] == batch:
            return tensor.unsqueeze(1).contiguous()
        if tensor.shape[0] == 1 and batch != 1:
            tensor = tensor.expand(batch, -1, -1)
            return tensor.unsqueeze(1).contiguous()
        if tensor.shape == (height, width):
            tensor = tensor.unsqueeze(0).expand(batch, -1, -1)
            return tensor.unsqueeze(1).contiguous()
    elif tensor.ndim == 2 and tensor.shape == (height, width):
        tensor = tensor.unsqueeze(0).unsqueeze(0).expand(batch, 1, -1, -1)
        return tensor.contiguous()

    raise ValueError("Spatial map must broadcast to shape (B, 1, H, W).")


def _forward_splat_impl(
    image: "torch.Tensor",
    flow: "torch.Tensor",
    weight_map: "torch.Tensor",
) -> "torch.Tensor":
    """Perform weighted forward splatting in ``(B, C, H, W)`` layout.

    Parameters
    ----------
    image : torch.Tensor
        Image tensor in ``(B, C, H, W)`` layout.
    flow : torch.Tensor
        Flow tensor in ``(B, 2, H, W)`` layout.
    weight_map : torch.Tensor
        Per-pixel weights with shape ``(B, 1, H, W)``.

    Returns
    -------
    torch.Tensor
        Forward-splatted image in ``(B, C, H, W)`` layout.
    """
    batch, channels, height, width = image.shape
    image_flat = image.view(batch, channels, height * width)

    device = image.device
    dtype = image.dtype

    yy, xx = torch.meshgrid(
        torch.arange(height, device=device, dtype=dtype),
        torch.arange(width, device=device, dtype=dtype),
        indexing="ij",
    )
    xx = xx.unsqueeze(0).expand(batch, -1, -1)
    yy = yy.unsqueeze(0).expand(batch, -1, -1)

    flow_x = flow[:, 0, :, :]
    flow_y = flow[:, 1, :, :]

    target_x = xx + flow_x
    target_y = yy + flow_y

    x0 = torch.floor(target_x)
    y0 = torch.floor(target_y)
    x1 = x0 + 1
    y1 = y0 + 1

    fx = target_x - x0
    fy = target_y - y0

    def _mask(x_coord, y_coord):
        return (
            (x_coord >= 0)
            & (x_coord <= width - 1)
            & (y_coord >= 0)
            & (y_coord <= height - 1)
        ).to(dtype)

    base_weight = weight_map.squeeze(1)

    mask00 = _mask(x0, y0)
    mask10 = _mask(x1, y0)
    mask01 = _mask(x0, y1)
    mask11 = _mask(x1, y1)

    w00 = (1 - fx) * (1 - fy) * base_weight * mask00
    w10 = fx * (1 - fy) * base_weight * mask10
    w01 = (1 - fx) * fy * base_weight * mask01
    w11 = fx * fy * base_weight * mask11

    x0i = x0.clamp(0, width - 1).long()
    x1i = x1.clamp(0, width - 1).long()
    y0i = y0.clamp(0, height - 1).long()
    y1i = y1.clamp(0, height - 1).long()

    out = torch.zeros((batch, channels, height * width), device=device, dtype=dtype)
    den = torch.zeros((batch, 1, height * width), device=device, dtype=dtype)

    def _splat(weights, x_idx, y_idx):
        if torch.count_nonzero(weights) == 0:
            return
        linear_idx = (y_idx * width + x_idx).view(batch, 1, -1)
        weights_flat = weights.view(batch, 1, -1)
        expanded_idx = linear_idx.expand(-1, channels, -1)

        out.scatter_add_(2, expanded_idx, image_flat * weights_flat)
        den.scatter_add_(2, linear_idx, weights_flat)

    for weights, x_idx, y_idx in (
        (w00, x0i, y0i),
        (w10, x1i, y0i),
        (w01, x0i, y1i),
        (w11, x1i, y1i),
    ):
        _splat(weights, x_idx, y_idx)

    den = torch.where(den == 0, torch.ones_like(den), den)
    return (out / den).view(batch, channels, height, width)


def forward_splat(image, flow, weight=None):
    """Forward warp an image using bilinear splatting.

    Parameters
    ----------
    image : array-like or torch.Tensor
        Image in any supported layout, e.g. ``(B, C, H, W)``, ``(C, H, W)``,
        ``(H, W, C)`` or ``(H, W)``.
    flow : array-like or torch.Tensor
        Flow field with layout ``(B, 2, H, W)`` or ``(H, W, 2)`` describing the
        forward displacement per pixel.
    weight : array-like or torch.Tensor, optional
        Additional per-pixel weights broadcastable to ``(B, 1, H, W)``. When
        omitted, uniform weights are used.

    Returns
    -------
    Same type as ``image``
        Forward warped image converted back to the original layout and dtype.

    Raises
    ------
    ImportError
        If ``torch`` is not available.
    ValueError
        If image and flow batch or spatial dimensions do not align.
    """
    if not HAS_TORCH:
        raise ImportError("torch is required for forward splatting.")

    input_is_tensor = torch.is_tensor(image)
    input_is_numpy = isinstance(image, np.ndarray)
    image_tensor = image if input_is_tensor else torch.as_tensor(image)
    flow_tensor = flow if torch.is_tensor(flow) else torch.as_tensor(flow)

    orig_dtype = image_tensor.dtype
    orig_device = image_tensor.device if input_is_tensor else None
    orig_numpy_dtype = image.dtype if input_is_numpy else None

    common_device = flow_tensor.device
    if image_tensor.device != common_device:
        if common_device.type == "cpu":
            common_device = image_tensor.device
        else:
            image_tensor = image_tensor.to(device=common_device)
    else:
        common_device = image_tensor.device

    compute_dtype = torch.promote_types(image_tensor.dtype, torch.float32)
    image_tensor = image_tensor.to(device=common_device, dtype=compute_dtype)
    flow_tensor = flow_tensor.to(device=common_device, dtype=compute_dtype)

    image_bchw, image_layout = _to_bchw(image_tensor)
    flow_b2hw, _ = _to_b2hw(flow_tensor)

    if image_bchw.shape[0] != flow_b2hw.shape[0]:
        if image_bchw.shape[0] == 1:
            image_bchw = image_bchw.expand(flow_b2hw.shape[0], -1, -1, -1)
        elif flow_b2hw.shape[0] == 1:
            flow_b2hw = flow_b2hw.expand(image_bchw.shape[0], -1, -1, -1)
        else:
            raise ValueError("Image and flow batch sizes must match.")

    if image_bchw.shape[2:] != flow_b2hw.shape[2:]:
        raise ValueError("Image and flow spatial dimensions must align.")

    weight_map = _prepare_spatial_map(
        weight,
        image_bchw.shape[0],
        image_bchw.shape[2],
        image_bchw.shape[3],
        image_bchw.device,
        image_bchw.dtype,
    )

    result = _forward_splat_impl(image_bchw, flow_b2hw, weight_map)
    result = _from_bchw(result, image_layout)

    if input_is_tensor:
        result = result.to(dtype=orig_dtype)
        if orig_device is not None:
            result = result.to(device=orig_device)
        return result
    if input_is_numpy:
        np_result = result.cpu().numpy()
        if orig_numpy_dtype is not None:
            np_result = np_result.astype(orig_numpy_dtype, copy=False)
        return np_result
    return result


def softmax_splat(image, flow, importance, temperature=1.0, clamp=50.0):
    """Forward warp an image using softmax splatting (Niklaus et al., 2020).

    Parameters
    ----------
    image : array-like or torch.Tensor
        Image tensor in any layout supported by :func:`forward_splat`.
    flow : array-like or torch.Tensor
        Flow field with layout ``(B, 2, H, W)`` or ``(H, W, 2)`` describing the
        forward displacement per pixel.
    importance : array-like or torch.Tensor
        Importance or occlusion map broadcastable to ``(B, 1, H, W)`` that
        controls blending weights.
    temperature : float, optional
        Softmax temperature. Higher values sharpen the aggregation and lower
        values smooth the blend. Default is ``1.0``.
    clamp : float, optional
        Absolute value used to clamp the exponent for numerical stability.
        Default is ``50.0``.

    Returns
    -------
    Same type as ``image``
        Forward warped image aggregated with softmax-normalised weights.

    Raises
    ------
    ImportError
        If ``torch`` is not available.
    ValueError
        If image and flow batch or spatial dimensions do not align.
    """
    if not HAS_TORCH:
        raise ImportError("torch is required for softmax splatting.")

    input_is_tensor = torch.is_tensor(image)
    input_is_numpy = isinstance(image, np.ndarray)
    image_tensor = image if input_is_tensor else torch.as_tensor(image)
    flow_tensor = flow if torch.is_tensor(flow) else torch.as_tensor(flow)

    orig_dtype = image_tensor.dtype
    orig_device = image_tensor.device if input_is_tensor else None
    orig_numpy_dtype = image.dtype if input_is_numpy else None

    common_device = flow_tensor.device
    if image_tensor.device != common_device:
        if common_device.type == "cpu":
            common_device = image_tensor.device
        else:
            image_tensor = image_tensor.to(device=common_device)
    else:
        common_device = image_tensor.device

    compute_dtype = torch.promote_types(image_tensor.dtype, torch.float32)
    image_tensor = image_tensor.to(device=common_device, dtype=compute_dtype)
    flow_tensor = flow_tensor.to(device=common_device, dtype=compute_dtype)

    image_bchw, image_layout = _to_bchw(image_tensor)
    flow_b2hw, _ = _to_b2hw(flow_tensor)

    if image_bchw.shape[0] != flow_b2hw.shape[0]:
        if image_bchw.shape[0] == 1:
            image_bchw = image_bchw.expand(flow_b2hw.shape[0], -1, -1, -1)
        elif flow_b2hw.shape[0] == 1:
            flow_b2hw = flow_b2hw.expand(image_bchw.shape[0], -1, -1, -1)
        else:
            raise ValueError("Image and flow batch sizes must match.")

    if image_bchw.shape[2:] != flow_b2hw.shape[2:]:
        raise ValueError("Image and flow spatial dimensions must align.")

    importance_map = _prepare_spatial_map(
        importance,
        image_bchw.shape[0],
        image_bchw.shape[2],
        image_bchw.shape[3],
        image_bchw.device,
        image_bchw.dtype,
    )

    scaled = torch.clamp(importance_map * temperature, min=-clamp, max=clamp)
    weight_map = torch.exp(scaled)

    result = _forward_splat_impl(image_bchw, flow_b2hw, weight_map)
    result = _from_bchw(result, image_layout)

    if input_is_tensor:
        result = result.to(dtype=orig_dtype)
        if orig_device is not None:
            result = result.to(device=orig_device)
        return result
    if input_is_numpy:
        np_result = result.cpu().numpy()
        if orig_numpy_dtype is not None:
            np_result = np_result.astype(orig_numpy_dtype, copy=False)
        return np_result
    return result


class TorchForwardSplatWarper:
    """Torch implementation of forward and softmax splatting."""

    def __init__(
        self, mode: str = "average", temperature: float = 1.0, clamp: float = 50.0
    ):
        if not HAS_TORCH:
            raise ImportError("torch is required for TorchForwardSplatWarper.")
        if mode not in {"average", "softmax"}:
            raise ValueError("mode must be either 'average' or 'softmax'.")
        self.mode = mode
        self.temperature = temperature
        self.clamp = clamp

    def __call__(self, image, flow, *, importance=None, weight=None):
        if self.mode == "average":
            return forward_splat(image, flow, weight=weight)
        if importance is None:
            raise ValueError("importance map must be provided for softmax splatting.")
        return softmax_splat(
            image,
            flow,
            importance=importance,
            temperature=self.temperature,
            clamp=self.clamp,
        )


__all__ = [
    "HAS_MODERNGL",
    "HAS_TORCH",
    "warp_image_pc",
    "warp_image_pc_single",
    "warp_image_backwards",
    "OnlineFrameWarper",
    "forward_splat",
    "softmax_splat",
    "TorchForwardSplatWarper",
]



# ------------------------------------------------------------------------------------
# AUTHOR STATEMENT
#
# Author: Philipp Flotho (philipp.flotho[at]uni-saarland.de)
#
# For citation, please refer to the project README.
# If you believe any confidential or proprietary content is included, please notify me.
#
# Copyright (c) 2025, Philipp Flotho
# ------------------------------------------------------------------------------------

import timm
import torch
import torch.nn as nn
import cv2
from torchvision.transforms import Normalize

import numpy as np

from os.path import join
from neurovc.thermal_landmarks import TFWLandmarker

import requests
from tqdm import tqdm
import os
from pathlib import Path

MOBILENET = "mobilenetv2_100"
RESNET = "resnet101"

DEVICE = "cuda"


class _ModelDownloader:
    def __init__(self, model_name, save_dir="~/.neurovc/models"):
        self.model_name = model_name
        self.save_dir = Path(os.path.expanduser(save_dir))
        self.file_id = _file_id_map.get(model_name)
        if not self.file_id:
            raise ValueError(
                f"Model name '{model_name}' is not valid. Check the file_id_map."
            )
        self.model_url = (
            f"https://drive.google.com/uc?export=download&id={self.file_id}"
        )
        self.model_path = self.save_dir / f"{model_name}.pt"

    def download_model(self):
        self.save_dir.mkdir(parents=True, exist_ok=True)

        if self.model_path.exists():
            return self.model_path

        print(f"Downloading {self.model_name}...")
        response = requests.get(self.model_url, stream=True)
        if response.status_code == 200:
            total_size = int(response.headers.get("content-length", 0))
            with (
                open(self.model_path, "wb") as f,
                tqdm(
                    desc=f"Downloading {self.model_name}",
                    total=total_size,
                    unit="B",
                    unit_scale=True,
                    unit_divisor=1024,
                ) as bar,
            ):
                for chunk in response.iter_content(chunk_size=8192):
                    f.write(chunk)
                    bar.update(len(chunk))
            print(f"Model downloaded to {self.model_path}")
        else:
            raise Exception(f"Failed to download model: {response.status_code}")
        return self.model_path


def _get_model(model_name="478"):
    downloader = _ModelDownloader(model_name)
    weights_path = downloader.download_model()
    model_path_final_joint = join(os.path.dirname(weights_path), "joint_converted.pt")
    convert_model(
        weights_path, model_path_final_joint, n_landmarks=int(model_name), mode="JOINT"
    )
    return model_path_final_joint


def _warping_depth(eta, levels, m, n):
    min_dim = min(m, n)
    warping_depth = 0
    d = warping_depth

    for i in range(levels):
        warping_depth += 1
        min_dim = min_dim * eta
        if round(min_dim) < 224:
            break
        d = warping_depth
    return d


def convert_model(model_path_in, model_path_out, mode="", n_landmarks=70):
    model = DMMv2(n_landmarks=n_landmarks)
    dmm = nn.DataParallel(model, device_ids=[0])
    dmm.load_state_dict(
        torch.load(model_path_in, map_location=torch.device("cpu"), weights_only=True),
        strict=False,
    )
    torch.save(dmm.module.state_dict(), model_path_out)


def create_model(model):
    model = timm.create_model(model, pretrained=True)
    model.classifier.requires_grad = False

    def forward_features(x):
        x = model.forward_features(x)
        return model.global_pool(x)

    return model, forward_features


class DMMv2(nn.Module):
    def __init__(self, n_landmarks, use_depth=False):
        super().__init__()
        self.use_depth = False
        self.feature_network = timm.create_model(
            "mobilenetv2_100", pretrained=True, num_classes=0
        )
        self.feature_network.classifier.requires_grad = False

        self.n_landmarks = n_landmarks
        self.fc = nn.Linear(1280, n_landmarks * (self.use_depth + 3))
        nn.init.xavier_uniform_(self.fc.weight)

        self.fc.requires_grad_(True)
        self.act = nn.ReLU()
        self.act_leaky = nn.LeakyReLU()

    def forward(self, x):
        x = self.feature_network.forward_features(x)
        x = self.feature_network.global_pool(x)
        x = self.fc(x)
        x = x.reshape((x.shape[0], self.n_landmarks, self.use_depth + 3))
        x = self.act_leaky(x)

        return x


class ThermalLandmarks:
    def __init__(
        self,
        model_path=None,
        device="cpu",
        gpus=[0, 1],
        eta=0.75,
        max_lvl=0,
        stride=100,
        n_landmarks=478,
        normalize=True,
    ):
        self.face_tracker = TFWLandmarker()

        dmm = DMMv2(n_landmarks=n_landmarks)
        self.n_landmarks = n_landmarks

        model_path = _get_model(str(n_landmarks))
        print(model_path)
        dmm.load_state_dict(torch.load(model_path, weights_only=True), strict=False)

        if device == "cuda":
            dmm = nn.DataParallel(dmm, device_ids=gpus)
        dmm.eval()
        dmm.to(device)
        self.device = device
        self.dmm = dmm
        self.img_shape = None
        self.eta = eta
        self.max_lvl = max_lvl
        self.stride = stride
        self.last_sparse_lm = None
        if normalize:
            tmp = Normalize(
                mean=torch.tensor([0.4850, 0.4560, 0.4060]),
                std=torch.tensor([0.2290, 0.2240, 0.2250]),
            )
            self.transform = lambda x: tmp(x / 255.0)
        else:
            self.transform = lambda x: x

    def process(self, image, sliding_window=False, multi=False):
        if self.img_shape is None:
            img_shape = image.shape[:2]
            wp = _warping_depth(self.eta, 100, *img_shape)

        if len(image.shape) == 2:
            image = image.astype(float)
            self.last_sparse_lm = self.face_tracker.detect(image)
            image = np.clip(image, 20, 40)
            image -= image.min()
            image /= image.max()
            image = np.stack([image, image, image], 2) * 255
        else:
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        if not sliding_window:
            if multi:
                return self.get_landmarks_multi(image)
            return self.get_landmarks_single(image)

        best_score = np.inf
        lm = None
        best_scores = None
        for i in range(wp, self.max_lvl - 1, -1):
            lvl_factor = self.eta**i
            size = (
                int(round(img_shape[1] * lvl_factor)),
                int(round(img_shape[0] * lvl_factor)),
            )
            img = cv2.resize(image, size)
            if len(img.shape) == 2:
                img = np.expand_dims(img, 2)
            hx = img_shape[1] / size[0]
            hy = img_shape[0] / size[1]
            lm_lvl, score, scores = self.get_landmarks(
                img.astype(np.float32), stride=self.stride
            )
            if score < best_score:
                best_score = score
                best_scores = scores
                lm = lm_lvl * np.expand_dims(np.array([hx, hy]), 0)

        return lm, best_scores

    def _refine_landmarks(self, img, lm_scaled):
        x_coords = lm_scaled[:, 0]
        y_coords = lm_scaled[:, 1]

        y_min = np.min(y_coords)
        y_max = np.max(y_coords)
        x_min = np.min(x_coords)
        x_max = np.max(x_coords)

        largest_side = max(y_max - y_min, x_max - x_min) * 1.25
        y_center = (y_max + y_min) / 2
        x_center = (x_max + x_min) / 2

        padding = int(largest_side // 2)

        padded_img = cv2.copyMakeBorder(
            img,
            padding,
            padding,
            padding,
            padding,
            cv2.BORDER_CONSTANT,
            value=[0, 0, 0],
        )

        x_start = int(x_center - largest_side / 2 + padding)
        y_start = int(y_center - largest_side / 2 + padding)
        x_end = int(x_center + largest_side / 2 + padding)
        y_end = int(y_center + largest_side / 2 + padding)

        patch = padded_img[y_start:y_end, x_start:x_end]

        x = cv2.resize(patch, (224, 224))
        x = (
            torch.from_numpy(x)
            .to(torch.float32)
            .to(self.device)
            .permute(2, 0, 1)
            .unsqueeze(0)
        )

        with torch.no_grad():
            cropped_transformed = self.transform(x)
            refined_lm = self.dmm(cropped_transformed)

        refined_lm_scaled = (
            (refined_lm[..., :-1] * largest_side).cpu().detach().squeeze().numpy()
        )
        refined_lm_scaled += np.array([[x_start - padding, y_start - padding]])
        confidences = refined_lm[..., -1].cpu().detach().squeeze().numpy()
        return refined_lm_scaled, confidences

    def get_landmarks_single(self, img):
        results = self.last_sparse_lm
        if len(results) == 0:
            return -np.ones((self.n_landmarks, 2)), np.zeros(self.n_landmarks)
        lm_scaled = results[0]["landmarks"]
        lm_scaled = np.array(lm_scaled).reshape((-1, 2))

        bbox = results[0]["box"]
        bbox = np.array(bbox).reshape((-1, 2))
        if lm_scaled is None:
            raise ValueError("Use mediapipe for dense RGB landmarks!")
        if -1 in lm_scaled:
            return lm_scaled, np.zeros(lm_scaled.shape[0])
        lm_scaled, confidences = self._refine_landmarks(img, bbox)

        return lm_scaled, confidences

    def get_landmarks_multi(self, img):
        results = self.last_sparse_lm
        if len(results) == 0:
            return [], []
        landmarks = []
        confidences_all = []
        for result in results:
            lm_scaled = result["landmarks"]
            lm_scaled = np.array(lm_scaled).reshape((-1, 2))
            bbox = result["box"]
            bbox = np.array(bbox).reshape((-1, 2))
            if lm_scaled is None:
                raise ValueError("Use mediapipe for dense RGB landmarks!")
            if -1 in lm_scaled:
                landmarks.append(-np.ones((self.n_landmarks, 2)))
                confidences_all.append(np.zeros(self.n_landmarks))
                continue
            refined, confidences = self._refine_landmarks(img, bbox)
            landmarks.append(refined)
            confidences_all.append(confidences)
        return landmarks, confidences_all

    def get_landmarks(self, img, stride=50, refine=True):
        """Sliding window implementation for the landmarks"""

        img_dims = img.shape
        y_pad = 224 - img_dims[0] % 224
        x_pad = 224 - img_dims[1] % 224

        y_pad_l = y_pad // 2
        y_pad_r = y_pad // 2 + y_pad % 2

        x_pad_l = x_pad // 2
        x_pad_r = x_pad // 2 + y_pad % 2

        pad = (0, 0, x_pad_l, x_pad_r, y_pad_l, y_pad_r)

        with torch.no_grad():
            x = torch.nn.functional.pad(
                torch.from_numpy(img).to(torch.float32), pad
            ).to(self.device)
            img_unfold = x.unfold(0, 224, stride).unfold(1, 224, stride)
            s = img_unfold.shape
            img_unfold = img_unfold.reshape((s[0] * s[1],) + s[2:])
            lm = self.dmm(self.transform(img_unfold))

            best_scores = lm[..., -1].mean(1)
            best_score_idx = best_scores.argmin().item()
            best_score = best_scores[best_score_idx].item()
            offset = (
                torch.Tensor(
                    [
                        stride * (best_score_idx % s[1]),
                        stride * (best_score_idx // s[1]),
                    ]
                )
                .unsqueeze(0)
                .to(self.device)
            )
            lm = lm[best_score_idx]
            lm_out = lm[:, :-1] * 224 + offset
            lm_out = lm_out - torch.tensor([x_pad_l, y_pad_l]).unsqueeze(0).to(
                self.device
            )

        lm_scaled = lm_out.cpu().detach().numpy()
        confidences = lm[..., -1].cpu().detach().numpy()

        return lm_scaled, best_score, confidences


_file_id_map = {
    "478": "1DZU3OOACp8gqxCxotZGwe3_gyNJCoY1p",
    "70": "1DqBVVmw9NscDELsnCxB4Pt9ltVUuNoWR",
}



import numpy as np
import pytest

from neurovc.momag import flow_processing as fp


def test_get_motion_magnitude_returns_expected_norm():
    flow = np.array([[[3.0, 4.0]]], dtype=np.float32)
    magnitudes = fp.get_motion_magnitude(flow)
    assert magnitudes.shape == (1, 1)
    assert np.allclose(magnitudes[0, 0], 5.0)


def test_compressive_function_thresh_scales_below_threshold():
    g_mag = np.array([[0.5, 1.5]], dtype=np.float32)
    result = fp.compressive_function_thresh(g_mag.copy(), alpha=2.0, threshold=1.0)
    expected = np.array([[1.0, 1.5]], dtype=np.float32)
    assert np.allclose(result, expected)


def test_const_compressor_scales_flow():
    flow = np.ones((1, 1, 2), dtype=np.float32)
    compressor = fp.ConstCompressor(alpha=2.0)
    out = compressor(flow.copy())
    g = np.sqrt(2.0)
    expected_value = (2.0 * g) / (g + 0.0001)
    assert np.allclose(out, expected_value)


def test_thresh_compressor_applies_threshold():
    flow = np.ones((1, 1, 2), dtype=np.float32)
    compressor = fp.ThreshCompressor(alpha=3.0, threshold=2.0)
    out = compressor(flow.copy())
    g = np.sqrt(2.0)
    expected_value = (3.0 * g) / (g + 0.0001)
    assert np.allclose(out, expected_value)


def test_flow_decomposer_creates_mask_and_decomposes_zero_flow():
    landmarks = np.array(
        [
            [8.0, 8.0, -0.2],
            [24.0, 8.0, -0.1],
            [24.0, 24.0, 0.1],
            [8.0, 24.0, 0.3],
        ],
        dtype=np.float32,
    )
    decomposer = fp.FlowDecomposer(landmarks, (32, 32), [0, 1, 2, 3])
    assert decomposer.mask.shape == (32, 32)
    assert np.any(decomposer.mask > 0)

    flow = np.zeros((32, 32, 2), dtype=np.float32)
    global_flow, local_flow = decomposer.decompose(flow)
    assert np.allclose(global_flow, 0.0)
    assert np.allclose(local_flow, 0.0)


def test_alpha_looper_cycles_values():
    looper = fp.AlphaLooper(alpha=(0.0, 2.0), step=0.5)
    values = [looper() for _ in range(6)]
    assert values == [0.0, 0.5, 1.0, 1.5, 1.5, 1.0]
    looper.reset()
    assert looper() == 0.0


def test_extract_landmarks_success():
    class _DummyLM:
        def __init__(self, x, y, z):
            self.x = x
            self.y = y
            self.z = z

    class _DummyResults:
        def __init__(self):
            self.multi_face_landmarks = [
                type(
                    "LMList",
                    (),
                    {"landmark": [_DummyLM(0.1, 0.2, 0.3), _DummyLM(0.4, 0.5, 0.6)]},
                )
            ]

    result = fp._extract_landmarks(_DummyResults())
    expected = np.array([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]], dtype=float)
    assert np.allclose(result, expected)


def test_extract_landmarks_raises_without_faces():
    with pytest.raises(ValueError):
        fp._extract_landmarks(None)



"""Motion magnification helpers."""

from PIL import Image, ImageDraw
import cv2
import mediapipe as mp
import numpy as np
from numba import jit
from scipy.interpolate import griddata
from scipy.spatial import ConvexHull

import neurovc
import neurovc as nvc
from neurovc.momag.framewarpers import OnlineFrameWarper, warp_image_backwards
from neurovc.util.IO_util import CircularFrameBuffer

__author__ = "Philipp Flotho"


def _gray_augmentor(frame):
    return cv2.cvtColor(frame, cv2.COLOR_RGB2GRAY)


def _extract_landmarks(results):
    if results is None or results.multi_face_landmarks is None:
        raise ValueError("No landmarks available in Mediapipe results.")
    return np.array(
        [[lm.x, lm.y, lm.z] for lm in results.multi_face_landmarks[0].landmark],
        dtype=float,
    )


def _identity(frame):
    return frame


def get_motion_magnitude(w):
    """Compute the magnitude of a 2D flow field.

    Parameters
    ----------
    w : ndarray
        Flow field with shape ``(H, W, 2)``.

    Returns
    -------
    ndarray
        Magnitude map with shape ``(H, W)``.
    """
    return np.sqrt(w[:, :, 0] * w[:, :, 0] + w[:, :, 1] * w[:, :, 1])


def compressive_function(G_mag, alpha, beta):
    """Apply a compressive non-linearity to a magnitude map.

    Parameters
    ----------
    G_mag : ndarray
        Magnitude map.
    alpha : float
        Scaling factor.
    beta : float
        Compression exponent between 0 and 1.

    Returns
    -------
    ndarray
        Compressed magnitude map.
    """
    eps = 0.1
    beta = 1 - beta
    G_mag[G_mag < eps] = eps
    g_comp = (alpha / G_mag) * np.power((G_mag / alpha), beta)
    g_comp[g_comp < 1] = 1

    return g_comp


class MagnitudeCompressor:
    def __call__(self, w):
        g_mag = np.repeat(get_motion_magnitude(w)[:, :, None], 2, axis=2)
        w /= g_mag + 0.0001
        g_comp = self.comp(g_mag)
        return g_comp * w


class GradMagnitudeCompressor(MagnitudeCompressor):
    def __init__(self, **kwargs):
        self.comp = lambda x: compressive_function(x, **kwargs)


class ConstCompressor(MagnitudeCompressor):
    def __init__(self, alpha):
        self.comp = lambda x: alpha * x


def compressive_function_thresh(G_mag, alpha, threshold):
    """Scale magnitudes below ``threshold`` by ``alpha``.

    Parameters
    ----------
    G_mag : ndarray
        Magnitude map.
    alpha : float
        Multiplicative factor for values below ``threshold``.
    threshold : float
        Threshold separating scaled and untouched regions.

    Returns
    -------
    ndarray
        Adjusted magnitude map.
    """
    g_comp = G_mag
    g_comp[G_mag < threshold] *= alpha
    return g_comp


class ThreshCompressor(MagnitudeCompressor):
    def __init__(self, alpha, threshold):
        self.comp = lambda x: compressive_function_thresh(x, alpha, threshold)


@jit(nopython=True)
def diffusion_loop(img_out, idx, k, h=1):
    """Perform iterative diffusion over selected pixels.

    Parameters
    ----------
    img_out : ndarray
        Image to be diffused.
    idx : ndarray
        Array of pixel coordinates for diffusion updates.
    k : int
        Number of iterations.
    h : float, optional
        Diffusion step size. Default is ``1``.

    Returns
    -------
    ndarray
        Diffused image array.
    """
    for time in range(k):
        for i, j in idx:
            img_out[i, j] += (0.25 / (h * h)) * (
                img_out[i, j + 1]
                + img_out[i, j - 1]
                + img_out[i - 1, j]
                + img_out[i + 1, j]
                - 4 * img_out[i, j]
            )

    return img_out


class FlowDecomposer:
    """Split dense flow into global and local components inside a mask."""

    def __init__(self, landmarks, dims, *idx):
        """
        Parameters
        ----------
        landmarks : ndarray
            Landmark coordinates with shape ``(N, 3)``.
        dims : tuple of int
            Flow dimensions as ``(height, width)``.
        *idx : iterable of int
            Landmark indices describing polygon regions.
        """
        self.idx = np.array(idx)
        self.dims = dims
        self.mask = None
        self.update_mask(landmarks)

    def update_mask(self, landmarks):
        """Recompute the diffusion mask from landmark polygons."""
        m, n = self.dims[0:2]
        img = Image.new("F", (n, m), 0.0)

        for idx in self.idx:
            points = landmarks[idx]
            hull = ConvexHull(points)

            args = np.argsort(hull.simplices, axis=0)
            _, idx = np.unique(hull.simplices[args[:, 0]][:, 0], return_index=True)

            points = [
                (x, y) for x, y, _ in points[hull.simplices[args[:, 0]][:, 0][idx]]
            ]
            ImageDraw.Draw(img).polygon(points, outline=255, fill=255)

        img = np.array(img)
        circle = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (1, 1))
        self.mask = cv2.dilate(img, circle, iterations=16)

    def decompose(self, flow):
        """Return global and local components of ``flow``."""
        mask_tmp = self.mask  # warp_image_pc_single(self.mask, flow)

        flow_global = np.empty(flow.shape, flow.dtype)

        flow_global[:, :, 0] = self.__diffusion(flow[:, :, 0], mask_tmp)
        flow_global[:, :, 1] = self.__diffusion(flow[:, :, 1], mask_tmp)

        flow_local = flow - flow_global

        return flow_global, flow_local

    def __flow_mag(self, flow, normalize=True):
        """Compute magnitude of ``flow`` optionally normalised to [0, 1]."""
        tmp = np.sqrt(
            np.multiply(flow[:, :, 0], flow[:, :, 0])
            + np.multiply(flow[:, :, 1], flow[:, :, 1])
        )
        if normalize:
            tmp /= 10.0
            tmp[tmp < 0.0] = 0.0
            tmp[tmp > 1.0] = 1.0

        return tmp

    def __diffusion(self, img, mask):
        """Diffuse ``img`` within ``mask`` to produce a smooth global flow."""
        idx = np.argwhere(mask != 0)

        img_out = np.empty(img.shape, img.dtype)
        np.copyto(img_out, img)
        img_out = img_out.astype(float)
        img_out[idx[:, 0], idx[:, 1]] = 0

        m, n = img_out.shape
        img_out_low = diffusion_loop(
            cv2.resize(img_out, (n // 8, m // 8)),
            np.argwhere(cv2.resize(mask, (n // 8, m // 8)) != 0),
            1000,
        )
        img_out[idx[:, 0], idx[:, 1]] = cv2.resize(img_out_low, (n, m))[
            idx[:, 0], idx[:, 1]
        ]
        img_out_low = diffusion_loop(
            cv2.resize(img_out, (n // 4, m // 4)),
            np.argwhere(cv2.resize(mask, (n // 4, m // 4)) != 0),
            1000,
        )
        img_out[idx[:, 0], idx[:, 1]] = cv2.resize(img_out_low, (n, m))[
            idx[:, 0], idx[:, 1]
        ]
        img_out[idx[:, 0], idx[:, 1]] = diffusion_loop(
            img_out, np.argwhere(mask != 0), 500
        )[idx[:, 0], idx[:, 1]]

        return img_out.astype(img.dtype)


class BasicMagnifier:
    """Base class encapsulating the optical-flow magnification pipeline."""

    def __init__(self, attenuation_function=None, augmentor=None, mesh_processor=None):
        """
        Parameters
        ----------
        attenuation_function : callable, optional
            Function applied to the flow magnitude before recombination.
        augmentor : callable, optional
            Preprocessing function applied to frames prior to flow estimation.
        mesh_processor : callable, optional
            Processor that extracts mesh or landmark information from frames.
        """
        self._augmentor = augmentor if augmentor is not None else _identity

        if attenuation_function is None:
            self._attenuation_function = ConstCompressor(5)
        else:
            self._attenuation_function = attenuation_function

        self.OF_inst = cv2.DISOpticalFlow.create(cv2.DISOpticalFlow_PRESET_MEDIUM)
        self.OF_inst.setUseSpatialPropagation(True)
        self.OF_inst.setFinestScale(1)
        self.OF_inst.setVariationalRefinementDelta(0.2)
        self.OF_inst.setVariationalRefinementGamma(0.8)
        self.OF_inst.setVariationalRefinementAlpha(0.5)
        self.OF_inst.setVariationalRefinementIterations(5)
        self.mesh_processor = mesh_processor

        self._init = False
        self.ref = None
        self.results_ref = None
        self.last_flow = None

    def _do_init(self, ref):
        """Initialise the internal frame warper based on ``ref``."""
        m, n = ref.shape[0:2]
        self.framewarper = OnlineFrameWarper((m, n))
        self._init = True

    def update_reference(self, ref, landmarks=None):
        """Update the reference frame and optional landmark information."""
        self.ref = ref
        self.results_ref = landmarks
        if landmarks is None and self.mesh_processor is not None:
            self.mesh_processor(ref)
            self.results_ref = self.mesh_processor.last_result[0]

        if not self._init:
            self._do_init(ref)

    def get_flow(self, frame):
        """Compute dense optical flow between the reference and ``frame``."""
        w = self.OF_inst.calc(
            self._augmentor(self.ref), self._augmentor(frame), self.last_flow
        )
        self.last_flow = w
        return w

    def __call__(self, frame):
        """Update internal state using ``frame`` and return it unchanged."""
        if not self._init:
            self.update_reference(frame)
        _ = self.get_flow(frame)
        return frame


class AlphaLooper:
    """Iterate through magnification factors to animate motion magnification.

    The values loop forward and backward across the specified range so repeated
    calls produce a smooth oscillation that can, for example, be used to render
    a video sequence from a fixed reference frame and its magnified variant.
    """

    def __init__(self, alpha, step=0.5):
        """
        Parameters
        ----------
        alpha : tuple of float
            Start and end values of the magnitude range.
        step : float
            Increment between consecutive magnitude values.
        """
        self.alpha = np.arange(alpha[0], alpha[1], step)
        self.idx = 0
        self.__forward = True

    def __call__(self):
        """Return the next magnification factor in the alternating sequence."""
        if self.__forward:
            alpha = self.alpha[self.idx]
        else:
            alpha = self.alpha[len(self.alpha) - self.idx - 1]
        self.idx = (self.idx + 1) % len(self.alpha)
        self.__forward = not self.__forward if self.idx == 0 else self.__forward
        return alpha

    def reset(self):
        """Reset the iterator to the start of the sequence."""
        self.idx = 0
        self.__forward = True


class MagnificationTask:
    """High-level task managing motion magnification with landmark cues."""

    def __init__(self):
        """Initialise buffers, optical flow magnifiers, and face mesh."""
        self.buffer = CircularFrameBuffer(10)
        self.last_local_flow = None
        self.last_global_flow = None
        self.last_frame = None
        self.last_global_depth = None

        magnifiers = [
            OnlineGlobalMagnifier(augmentor=_gray_augmentor),
            OnlineLandmarkMagnifier(nvc.LM_EYE_LEFT, augmentor=_gray_augmentor),
            OnlineLandmarkMagnifier(nvc.LM_EYE_RIGHT, augmentor=_gray_augmentor),
            OnlineLandmarkMagnifier(nvc.LM_FOREHEAD, augmentor=_gray_augmentor),
            OnlineLandmarkMagnifier(
                landmarks=nvc.LM_MOUTH,
                augmentor=_gray_augmentor,
            ),
        ]

        self.magnifiers = {
            ord(str(i + 1)): magnifier for i, magnifier in enumerate(magnifiers)
        }
        self.current_magnifier = magnifiers[0]

        mp_face_mesh = mp.solutions.face_mesh
        self.face_mesh = mp_face_mesh.FaceMesh(
            static_image_mode=True,
            refine_landmarks=True,
            max_num_faces=1,
            min_detection_confidence=0.5,
        )

        self.ref = None
        self.results_ref = None

    def __call__(self, frame):
        """Process ``frame`` to update reference buffers and flow state."""
        buf = self.buffer.get_oldest()
        results_frame = self.face_mesh.process(frame)

        self.results_frame = results_frame
        self.last_frame = frame

        if buf is None:
            self.buffer.append((frame, results_frame))
            self.ref = frame
            self.results_ref = results_frame
            return frame
        ref, results_ref = buf
        self.buffer.append((frame, results_frame))
        try:
            self.current_magnifier.update_reference(ref, results_ref)
            self.ref = ref
            self.results_ref = results_ref
            self.update_flow()
        except Exception:
            print("No Landmarks in reference.")

        if results_frame.multi_face_landmarks is None:
            return

        try:
            points = _extract_landmarks(results_ref)
        except Exception:
            return
        image_size = frame.shape[:2]
        points[:, 0] *= image_size[1]
        points[:, 1] *= image_size[0]
        grid_y, grid_x = np.mgrid[0 : image_size[0], 0 : image_size[1]]

        global_depth = griddata(
            points[:, :2], points[:, 2], (grid_x, grid_y), method="linear"
        )
        gd_nan_mask = np.isnan(global_depth)
        global_depth[gd_nan_mask] = global_depth[~gd_nan_mask].max()
        global_depth -= global_depth.min()
        global_depth /= global_depth.max()
        global_depth = 1 - global_depth
        cv2.imshow("depth", global_depth)
        self.last_global_depth = global_depth

    def get_mag(self, alpha):
        """Return the most recent frame warped with magnification ``alpha``."""
        if self.last_global_flow is None or self.last_local_flow is None:
            return self.last_frame
        flow = self.last_global_flow + self.last_local_flow
        return self.current_magnifier.framewarper.warp_image_uv(
            warp_image_backwards(self.last_frame, flow),
            self.last_global_flow + alpha * self.last_local_flow,
            self.last_global_depth,
        )

    def update_flow(self):
        """Refresh cached global and local flow components."""
        try:
            flow_global, flow_local = self.current_magnifier(
                self.last_frame, self.results_frame
            )
            self.last_global_flow = flow_global
            self.last_local_flow = flow_local
        except Exception:
            pass

    def set_magnifier(self, key, update_flow=False):
        """Switch the active magnifier using a numeric keyboard key."""
        key = key & 0xFF
        if key in self.magnifiers.keys():
            self.current_magnifier = self.magnifiers[key]
            try:
                self.current_magnifier.update_reference(self.ref, self.results_ref)
            except Exception:
                pass
            if update_flow:
                self.update_flow()


class OnlineGlobalMagnifier(BasicMagnifier):
    def __init__(self, augmentor=None):
        super().__init__(augmentor=augmentor)

    def requires_landmarks(self):
        return False

    def __call__(self, frame, landmarks=None):
        _ = super().__call__(frame)
        return np.zeros_like(self.last_flow), self.last_flow


class OnlineLandmarkMagnifier(BasicMagnifier):
    def __init__(
        self,
        landmarks=neurovc.LM_MOUTH,
        alpha=15,
        reference=None,
        attenuation_function=None,
        augmentor=None,
    ):
        super().__init__(
            attenuation_function=attenuation_function,
            mesh_processor=FacialMeshProcessor(),
            augmentor=augmentor,
        )

        mp_drawing = mp.solutions.drawing_utils
        mp_face_mesh = mp.solutions.face_mesh
        self.drawing_spec = mp_drawing.DrawingSpec(
            thickness=1, circle_radius=0, color=(0, 0, 0)
        )

        self.face_mesh = mp_face_mesh.FaceMesh(
            static_image_mode=True, max_num_faces=1, min_detection_confidence=0.5
        )

        self._init = False
        self.alpha = alpha

        self.framewarper = None
        self.decomposer = None
        self.ref = None
        self.landmarks = landmarks
        if reference is not None:
            self.update_reference(reference)

    def requires_landmarks(self):
        return True

    def update_reference(self, ref, landmarks=None):
        super().update_reference(ref)
        m, n = ref.shape[0:2]
        self.decomposer = FlowDecomposer(
            self.mesh_processor.get_last_landmarks(),
            (m, n),
            self.mesh_processor.get_region_from_id(self.landmarks),
        )

    def __call__(self, frame, landmarks=None):
        _ = super().__call__(frame)
        w = self.last_flow
        w_global, w_local = self.decomposer.decompose(w)
        return w_global, w_local

    def get_reference(self):
        mp_drawing = mp.solutions.drawing_utils
        mp_face_mesh = mp.solutions.face_mesh

        lm_viz = self.ref.copy()
        try:
            mp_drawing.draw_landmarks(
                image=lm_viz,
                landmark_list=self.mesh_processor.last_result[0][0],
                connections=mp_face_mesh.FACEMESH_TESSELATION,
                landmark_drawing_spec=self.drawing_spec,
                connection_drawing_spec=self.drawing_spec,
            )
        except Exception:
            pass
        for x, y, _ in self.mesh_processor.get_last_landmarks()[
            self.mesh_processor.get_region_from_id(self.landmarks)
        ]:
            cv2.circle(lm_viz, (int(np.round(x)), int(np.round(y))), 5, (0, 0, 255), -1)

        return lm_viz, self.ref


class FacialMeshProcessor:
    mouth_boundary_inner = [
        167,
        164,
        393,
        391,
        322,
        410,
        287,
        273,
        335,
        406,
        313,
        18,
        83,
        182,
        106,
        43,
        57,
        186,
        92,
        165,
    ]
    mouth_boundary = [
        203,
        98,
        97,
        2,
        326,
        327,
        423,
        426,
        436,
        432,
        422,
        424,
        418,
        421,
        200,
        201,
        194,
        204,
        202,
        216,
        212,
        206,
    ]
    mouth_boundary_outer = [
        203,
        98,
        97,
        2,
        326,
        327,
        423,
        425,
        427,
        430,
        434,
        431,
        207,
        205,
        214,
        210,
        211,
        418,
        421,
        200,
        201,
        194,
    ]
    eye_left_boundary_inner = [
        225,
        224,
        223,
        222,
        221,
        189,
        244,
        233,
        232,
        231,
        230,
        229,
        228,
        31,
        226,
        113,
    ]
    eye_right_boundary_inner = [
        464,
        413,
        441,
        442,
        443,
        444,
        445,
        342,
        446,
        448,
        449,
        450,
        451,
        452,
        453,
        261,
    ]
    eye_left_boundary = [
        46,
        53,
        52,
        65,
        55,
        193,
        245,
        128,
        121,
        120,
        119,
        118,
        117,
        111,
        35,
        124,
    ]
    eye_right_boundary = [
        285,
        417,
        465,
        357,
        350,
        349,
        348,
        347,
        346,
        340,
        265,
        353,
        276,
        283,
        282,
        295,
    ]
    forehead_temp_boundary = [
        109,
        69,
        66,
        65,
        55,
        8,
        285,
        295,
        282,
        334,
        333,
        332,
        297,
        338,
        10,
    ]
    _lms = [
        mouth_boundary,
        eye_left_boundary,
        eye_right_boundary,
        forehead_temp_boundary,
    ]

    def __init__(self):
        mp_drawing = mp.solutions.drawing_utils
        mp_face_mesh = mp.solutions.face_mesh
        self.drawing_spec = mp_drawing.DrawingSpec(
            thickness=1, circle_radius=0, color=(0, 0, 0)
        )

        self.face_mesh = mp_face_mesh.FaceMesh(
            static_image_mode=True, max_num_faces=1, min_detection_confidence=0.5
        )

        self.last_result = None

    def __call__(self, frame):
        results = self.face_mesh.process(frame)

        m, n = frame.shape[0:2]

        self.last_result = (results, (m, n))

    def get_region_from_id(self, id):
        ids = [1 & (id >> i) for i in range(4)]
        landmark_ids = []
        for i, b in enumerate(ids):
            if b:
                landmark_ids += self._lms[i]
        return landmark_ids

    def get_last_landmarks(self):
        results, (m, n) = self.last_result
        landmark_list = np.array(
            [
                [n * lm.x, m * lm.y, lm.z]
                for lm in results.multi_face_landmarks[0].landmark
            ]
        )
        return landmark_list


__all__ = [
    "MagnitudeCompressor",
    "GradMagnitudeCompressor",
    "ConstCompressor",
    "ThreshCompressor",
    "FlowDecomposer",
    "BasicMagnifier",
    "OnlineGlobalMagnifier",
    "OnlineLandmarkMagnifier",
    "FacialMeshProcessor",
    "AlphaLooper",
    "MagnificationTask",
]



from collections import deque
from threading import Thread
from time import time

try:
    import PySpin

    HAS_PYSPIN = True
except ModuleNotFoundError:
    PySpin = None  # type: ignore[assignment]
    HAS_PYSPIN = False

try:
    from screening_demo.stereo_config import GenericCamera as _ScreeningGenericCamera

    HAS_SCREENING_DEMO = True
except ModuleNotFoundError:
    HAS_SCREENING_DEMO = False
    from neurovc.util.camera_util import GenericCamera  # fallback stub
else:
    GenericCamera = _ScreeningGenericCamera
    del _ScreeningGenericCamera


class ThermalGrabber(Thread):
    def __init__(self, i=0):
        if not HAS_PYSPIN:
            raise ImportError(
                "PySpin is required for ThermalGrabber. Install the proprietary Spinnaker SDK to enable this feature."
            )
        Thread.__init__(self)
        self.__cam_id = i
        self.cam = None
        self.n_frame = 0
        self.__cam = None
        self.__callbacks = []
        self.frame_deque = deque()

    def run(self):
        # Thermal camera initialization (Spinnaker example):
        # Retrieve singleton reference to system object
        system = PySpin.System.GetInstance()

        # Get current library version
        version = system.GetLibraryVersion()
        print(
            "Library version: %d.%d.%d.%d"
            % (version.major, version.minor, version.type, version.build)
        )

        # Retrieve list of cameras from the system
        cam_list = system.GetCameras()

        num_cameras = cam_list.GetSize()

        print("Number of cameras detected: %d" % num_cameras)

        # if self.__cam_id > num_cameras:
        #    return

        self.__cam = cam_list[self.__cam_id]
        self.__cam.Init()

        nodemap_tldevice = self.__cam.GetTLDeviceNodeMap()
        nodemap = self.__cam.GetNodeMap()

        print("*** IMAGE ACQUISITION ***\n")
        try:
            # Set acquisition mode to continuous
            #
            #  *** NOTES ***
            #  Because the example acquires and saves 10 images, setting acquisition
            #  mode to continuous lets the example finish. If set to single frame
            #  or multiframe (at a lower number of images), the example would just
            #  hang. This would happen because the example has been written to
            #  acquire 10 images while the camera would have been programmed to
            #  retrieve less than that.
            #
            #  Setting the value of an enumeration node is slightly more complicated
            #  than other node types. Two nodes must be retrieved: first, the
            #  enumeration node is retrieved from the nodemap; and second, the entry
            #  node is retrieved from the enumeration node. The integer value of the
            #  entry node is then set as the new value of the enumeration node.
            #
            #  Notice that both the enumeration and the entry nodes are checked for
            #  availability and readability/writability. Enumeration nodes are
            #  generally readable and writable whereas their entry nodes are only
            #  ever readable.
            #
            #  Retrieve enumeration node from nodemap

            # In order to access the node entries, they have to be casted to a pointer type (CEnumerationPtr here)
            node_acquisition_mode = PySpin.CEnumerationPtr(
                nodemap.GetNode("AcquisitionMode")
            )
            node_temp_mode = PySpin.CEnumerationPtr(nodemap.GetNode("IRFormat"))
            if not PySpin.IsAvailable(node_acquisition_mode) or not PySpin.IsWritable(
                node_acquisition_mode
            ):
                print(
                    "Unable to set acquisition mode to continuous (enum retrieval). Aborting..."
                )
                return False

            # Retrieve entry node from enumeration node
            node_acquisition_mode_continuous = node_acquisition_mode.GetEntryByName(
                "Continuous"
            )
            node_temp_mode_linear = node_temp_mode.GetEntryByName(
                "TemperatureLinear10mK"
            )
            if not PySpin.IsAvailable(
                node_acquisition_mode_continuous
            ) or not PySpin.IsReadable(node_acquisition_mode_continuous):
                print(
                    "Unable to set acquisition mode to continuous (entry retrieval). Aborting..."
                )
                return False

            # Retrieve integer value from entry node
            temp_mode = node_temp_mode_linear.GetValue()
            acquisition_mode_continuous = node_acquisition_mode_continuous.GetValue()

            # Set integer value from entry node as new value of enumeration node
            node_temp_mode.SetIntValue(temp_mode)
            node_acquisition_mode.SetIntValue(acquisition_mode_continuous)

            print("Acquisition mode set to continuous...")

            #  Begin acquiring images
            #
            #  *** NOTES ***
            #  What happens when the camera begins acquiring images depends on the
            #  acquisition mode. Single frame captures only a single image, multi
            #  frame catures a set number of images, and continuous captures a
            #  continuous stream of images. Because the example calls for the
            #  retrieval of 10 images, continuous mode has been set.
            #
            #  *** LATER ***
            #  Image acquisition must be ended when no more images are needed.

            #  Retrieve device serial number for filename
            #
            #  *** NOTES ***
            #  The device serial number is retrieved in order to keep cameras from
            #  overwriting one another. Grabbing image IDs could also accomplish
            #  this.
            _device_serial_number = ""
            _node_device_serial_number = PySpin.CStringPtr(
                nodemap_tldevice.GetNode("DeviceSerialNumber")
            )
            self.__cam.BeginAcquisition()

            # for i in range(100):
            #    image_result = self.__cam.GetNextImage(1000)
            #    if not image_result.IsIncomplete():
            #        print("broken frame " + str(i))
            #    else:
            #        print("working frame " + str(i))

        except PySpin.SpinnakerException as ex:
            print("Error: %s" % ex)
            return False
        print("initialized")
        while True:
            # try:
            image_result = self.__cam.GetNextImage(1000)
            if not image_result.IsIncomplete():
                frame = image_result.GetNDArray()
                self.n_frame += 1
            else:
                continue
            # except:
            #    continue
            image_result.Release()
            ts = time() * 1000000

            # print("got frame " + str(self.n_frame))

            if frame is None:
                continue
            self.__callback(self.n_frame, ts, frame)
            result_tuple = (self.n_frame, ts, frame)
            self.frame_deque.append(result_tuple)

    def add_callback(self, callback):
        self.__callbacks.append(callback)

    def remove_callback(self, callback):
        self.__callbacks.remove(callback)

    def __callback(self, n_frame, ts, frame):
        for f in self.__callbacks:
            f(n_frame, ts, frame)


class ThermalCamera(GenericCamera):
    def __init__(self, i=0):
        if not HAS_PYSPIN:
            raise ImportError(
                "PySpin is required for ThermalCamera. Install the proprietary Spinnaker SDK to enable this feature."
            )
        self.__cam = None
        self.__cam_id = i

        self.n_frame = 0

    def start_acquisition(self):
        # Thermal camera initialization (Spinnaker example):
        # Retrieve singleton reference to system object
        system = PySpin.System.GetInstance()

        # Get current library version
        version = system.GetLibraryVersion()
        print(
            "Library version: %d.%d.%d.%d"
            % (version.major, version.minor, version.type, version.build)
        )

        # Retrieve list of cameras from the system
        cam_list = system.GetCameras()

        num_cameras = cam_list.GetSize()

        print("Number of cameras detected: %d" % num_cameras)

        # if self.__cam_id > num_cameras:
        #    return

        self.__cam = cam_list[self.__cam_id]
        self.__cam.Init()

        nodemap_tldevice = self.__cam.GetTLDeviceNodeMap()
        nodemap = self.__cam.GetNodeMap()

        print("*** IMAGE ACQUISITION ***\n")
        try:
            # Set acquisition mode to continuous
            #
            #  *** NOTES ***
            #  Because the example acquires and saves 10 images, setting acquisition
            #  mode to continuous lets the example finish. If set to single frame
            #  or multiframe (at a lower number of images), the example would just
            #  hang. This would happen because the example has been written to
            #  acquire 10 images while the camera would have been programmed to
            #  retrieve less than that.
            #
            #  Setting the value of an enumeration node is slightly more complicated
            #  than other node types. Two nodes must be retrieved: first, the
            #  enumeration node is retrieved from the nodemap; and second, the entry
            #  node is retrieved from the enumeration node. The integer value of the
            #  entry node is then set as the new value of the enumeration node.
            #
            #  Notice that both the enumeration and the entry nodes are checked for
            #  availability and readability/writability. Enumeration nodes are
            #  generally readable and writable whereas their entry nodes are only
            #  ever readable.
            #
            #  Retrieve enumeration node from nodemap

            # In order to access the node entries, they have to be casted to a pointer type (CEnumerationPtr here)
            node_acquisition_mode = PySpin.CEnumerationPtr(
                nodemap.GetNode("AcquisitionMode")
            )
            if not PySpin.IsAvailable(node_acquisition_mode) or not PySpin.IsWritable(
                node_acquisition_mode
            ):
                print(
                    "Unable to set acquisition mode to continuous (enum retrieval). Aborting..."
                )
                return False

            # Retrieve entry node from enumeration node
            node_acquisition_mode_continuous = node_acquisition_mode.GetEntryByName(
                "Continuous"
            )
            if not PySpin.IsAvailable(
                node_acquisition_mode_continuous
            ) or not PySpin.IsReadable(node_acquisition_mode_continuous):
                print(
                    "Unable to set acquisition mode to continuous (entry retrieval). Aborting..."
                )
                return False

            # Retrieve integer value from entry node
            acquisition_mode_continuous = node_acquisition_mode_continuous.GetValue()

            # Set integer value from entry node as new value of enumeration node
            node_acquisition_mode.SetIntValue(acquisition_mode_continuous)

            print("Acquisition mode set to continuous...")

            #  Begin acquiring images
            #
            #  *** NOTES ***
            #  What happens when the camera begins acquiring images depends on the
            #  acquisition mode. Single frame captures only a single image, multi
            #  frame catures a set number of images, and continuous captures a
            #  continuous stream of images. Because the example calls for the
            #  retrieval of 10 images, continuous mode has been set.
            #
            #  *** LATER ***
            #  Image acquisition must be ended when no more images are needed.

            #  Retrieve device serial number for filename
            #
            #  *** NOTES ***
            #  The device serial number is retrieved in order to keep cameras from
            #  overwriting one another. Grabbing image IDs could also accomplish
            #  this.
            _device_serial_number = ""
            _node_device_serial_number = PySpin.CStringPtr(
                nodemap_tldevice.GetNode("DeviceSerialNumber")
            )

            self.__cam.BeginAcquisition()

        except PySpin.SpinnakerException as ex:
            print("Error: %s" % ex)
            return False

    def get_image(self):
        try:
            image_result = self.__cam.GetNextImage(1)
            if not image_result.IsIncomplete():
                img = image_result.GetNDArray()
                img = img[:480, :640]
                self.n_frame += 1
                return self.n_frame, time(), img
            else:
                return None, None, None
        except Exception:
            return None, None, None

    def __del__(self):
        if self.__cam is not None:
            self.__cam.EndAcquisition()
        super(ThermalCamera, self).__del__()


__all__ = [
    "HAS_PYSPIN",
    "HAS_SCREENING_DEMO",
    "ThermalGrabber",
    "ThermalCamera",
]



from . import flow_processing as _flow_processing
from . import framewarpers as _framewarpers

HAS_MODERNGL = _framewarpers.HAS_MODERNGL
HAS_TORCH = _framewarpers.HAS_TORCH

__all__ = [
    "HAS_MODERNGL",
    "HAS_TORCH",
    *_flow_processing.__all__,
    *_framewarpers.__all__,
]

for _name in _flow_processing.__all__:
    globals()[_name] = getattr(_flow_processing, _name)

for _name in _framewarpers.__all__:
    globals()[_name] = getattr(_framewarpers, _name)

del _flow_processing, _framewarpers, _name



import numpy as np
import pytest

from neurovc.momag.framewarpers import (
    TorchForwardSplatWarper,
    forward_splat,
    softmax_splat,
)

torch = pytest.importorskip("torch")


def test_forward_splat_identity_tensor():
    image = torch.arange(9, dtype=torch.float32).view(1, 1, 3, 3)
    flow = torch.zeros(1, 2, 3, 3, dtype=torch.float32)

    warped = forward_splat(image, flow)

    assert warped.shape == image.shape
    assert warped.dtype == image.dtype
    assert torch.allclose(warped, image)


def test_forward_splat_identity_numpy():
    image = np.arange(16, dtype=np.float32).reshape(4, 4)
    flow = np.zeros((4, 4, 2), dtype=np.float32)

    warped = forward_splat(image, flow)

    assert isinstance(warped, np.ndarray)
    np.testing.assert_allclose(warped, image)


def test_forward_splat_respects_weight_map():
    image = torch.tensor([[[[1.0, 2.0], [3.0, 4.0]]]])
    flow = torch.zeros(1, 2, 2, 2)
    weight = torch.tensor([[[[0.0, 1.0], [1.0, 0.0]]]])

    warped = forward_splat(image, flow, weight=weight)

    expected = torch.tensor([[[[0.0, 2.0], [3.0, 0.0]]]])
    assert torch.allclose(warped, expected)


def test_forward_splat_preserves_dtype():
    image = torch.arange(4, dtype=torch.float64).view(1, 1, 2, 2)
    flow = torch.zeros(1, 2, 2, 2, dtype=torch.float64)

    warped = forward_splat(image, flow)

    assert warped.dtype == torch.float64


def test_softmax_splat_matches_forward_for_zero_importance():
    image = torch.rand(1, 3, 4, 5, dtype=torch.float32)
    flow = torch.zeros(1, 2, 4, 5, dtype=torch.float32)
    importance = torch.zeros(1, 1, 4, 5, dtype=torch.float32)

    avg_warp = forward_splat(image, flow)
    softmax_warp = softmax_splat(image, flow, importance)

    assert torch.allclose(avg_warp, softmax_warp, atol=1e-6)


def test_softmax_splat_numpy_input():
    image = np.random.rand(3, 5).astype(np.float32)
    flow = np.zeros((3, 5, 2), dtype=np.float32)
    importance = np.zeros((3, 5), dtype=np.float32)

    warped = softmax_splat(image, flow, importance)

    np.testing.assert_allclose(warped, image)


def test_torch_forward_splat_warper_average_mode():
    image = torch.rand(1, 1, 3, 3)
    flow = torch.zeros(1, 2, 3, 3)

    warper = TorchForwardSplatWarper(mode="average")
    warped = warper(image, flow)

    assert torch.allclose(warped, forward_splat(image, flow))


def test_torch_forward_splat_warper_softmax_requires_importance():
    image = torch.rand(1, 1, 2, 2)
    flow = torch.zeros(1, 2, 2, 2)

    warper = TorchForwardSplatWarper(mode="softmax")
    with pytest.raises(ValueError):
        warper(image, flow)


def test_torch_forward_splat_warper_softmax_mode():
    image = torch.tensor([[[[1.0, 0.0], [0.0, 0.0]]]])
    flow = torch.zeros(1, 2, 2, 2)
    importance = torch.tensor([[[[0.0, -10.0], [-10.0, -10.0]]]])

    warper = TorchForwardSplatWarper(mode="softmax", temperature=10.0)
    warped = warper(image, flow, importance=importance)

    assert torch.allclose(warped, image, atol=1e-6)
